/*
 * Copyright (c) 2003, The Regents of the University of California, through
 * Lawrence Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy). All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * (1) Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * (2) Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * 
 * (3) Neither the name of the University of California, Lawrence Berkeley
 * National Laboratory, U.S. Dept. of Energy nor the names of its contributors
 * may be used to endorse or promote products derived from this software without
 * specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 * 
 * You are under no obligation whatsoever to provide any bug fixes, patches, or
 * upgrades to the features, functionality or performance of the source code
 * ("Enhancements") to anyone; however, if you choose to make your Enhancements
 * available either publicly, or directly to Lawrence Berkeley National
 * Laboratory, without imposing a separate written license agreement for such
 * Enhancements, then you hereby grant the following license: a non-exclusive,
 * royalty-free perpetual license to install, use, modify, prepare derivative
 * works, incorporate into other computer software, distribute, and sublicense
 * such enhancements or derivative works thereof, in binary and source code
 * form.
 */
package nux.xom.xquery;


import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.URI;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.xml.transform.ErrorListener;
import javax.xml.transform.Source;
import javax.xml.transform.TransformerException;
import javax.xml.transform.URIResolver;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.sax.SAXSource;

import net.sf.saxon.Configuration;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.query.DynamicQueryContext;
import net.sf.saxon.query.StaticQueryContext;
import net.sf.saxon.query.XQueryExpression;
import net.sf.saxon.value.UntypedAtomicValue;
import net.sf.saxon.xom.DocumentWrapper;
import net.sf.saxon.xpath.XPathEvaluator;
import nu.xom.Attribute;
import nu.xom.Document;
import nu.xom.Element;
import nu.xom.Node;
import nu.xom.Nodes;
import nu.xom.ParentNode;
import nu.xom.ParsingException;
import nu.xom.Text;
import nu.xom.ValidityException;
import nu.xom.converters.DOMConverter;
import nux.xom.pool.DocumentURIResolver;
import nux.xom.pool.XOMUtil;

import org.xml.sax.InputSource;

/**
 * Compiled representation of a W3C XQuery (thread-safe).
 * Since XQuery is a superset of XPath 2.0 this class can also be used with 
 * plain XPath expressions as queries. Also note that XPath 2.0 supports regular expressions via
 * the <code>fn:matches</code>, <code>fn:replace</code>, and <code>fn:tokenize</code> functions.
 * <p> 
 * Instances are considered immutable and thread-safe; The same compiled query may 
 * be executed (evaluated) many times in series or in parallel,
 * just like {@link nu.xom.xslt.XSLTransform} objects.
 * <p>
 * Requires <a target="_blank" href="http://saxon.sourceforge.net/">saxon-8.1.1</a> or higher;
 * Does not care whether it is an open-source or commercial Saxon version.
 * To get started, simply add saxon8.jar to your classpath. 
 * Thanks to Michael Kay for that amazing piece of software!
 * In case of problems, first check the 
 * <a target="_blank" href="http://www.saxonica.com/documentation/documentation.html">
 * saxon documentation</a> and
 * search the <a target="_blank" href="http://lists.sourceforge.net/lists/listinfo/saxon-help">
 * saxon mailing list</a> as well as the 
 * <a target="_blank" href="http://marc.theaimsgroup.com/?l=xom-interest">
 * XOM mailing list</a>.
 * <p>
 * Example usage:
 * <pre>
 *     // find the links of all images in a XHTML-like document
 *     XQuery xquery = new XQuery("//*:img/@src", null);
 * 
 *     // find the links of all JPG images in a XHTML-like document via regular expression
 *     // XQuery xquery = new XQuery("//*:img/@src[matches(., '.jpg')]", null);
 * 
 *     Document doc = BuilderPool.GLOBAL_POOL.getBuilder(false).build(new File("/tmp/test.xml"));
 *     Nodes results = xquery.execute(doc).toNodes();
 *     for (int i=0; i < results.size(); i++) {
 *         System.out.println("node "+i+": "+results.get(i).toXML());
 *         //System.out.println("node "+i+": "+ XOMUtil.toPrettyXML(results.get(i)));
 *     }
 * </pre>
 *
 * Here is another typical relational join XQuery that, 
 * for all bicycles, lists the item number, description, and highest bid (if any), 
 * ordered by item number:
 * <pre>
 *     for $i in doc("items.xml")//item_tuple
 *     let $b := doc("bids.xml")//bid_tuple[itemno = $i/itemno]
 *     where contains($i/description, "Bicycle")
 *     order by $i/itemno
 *     return
 *        &lt;item_tuple>
 *           { $i/itemno }
 *           { $i/description }
 *           &lt;high_bid>{ max($b/bid) }&lt;/high_bid>
 *        &lt;/item_tuple>
 * </pre>
 * 
 * Here is a query (using namespaces) that selects all records that have a remark in German:
 * <pre>
 *     declare namespace music = "http://www.example.org/music/records";
 *     declare namespace xsd = "http://www.w3.org/2001/XMLSchema";
 *     declare base-uri "http://example.org";
 * 
 *     &lt;Q5 xmlns:music="http://www.example.org/music/records">
 *     {
 *        doc("tests/2003-11/usecases/ns/action.xml")//music:record[music:remark/@xml:lang = "de"]
 *     }
 *     &lt;/Q5>
 * </pre>
 * 
 * A query can declare external global variables, for example: 
 * <pre>
 *     declare variable $foo     as xs:string external (:: pragma saxon:default "hello" ::); 
 *     declare variable $size    as xs:integer external (:: pragma saxon:default 100 ::); 
 *     declare variable $myuri   as xs:anyURI external;
 *     declare variable $mynode  as node() external;
 *     declare variable $myelem  as element() external;
 *     declare variable $mynodes as node()* external;
 * </pre>
 * 
 * Such external global variable values can be passed to the query as follows:
 * <pre>
 *     Map vars = new HashMap();
 *     vars.put("foo", "hello world");
 *     vars.put("size", new Integer(99));
 *     vars.put("myuri", "http://www.w3.org/2001/XMLSchema");
 *     vars.put("mynode", new Document(new Element("xyz")));
 *     vars.put("myelem", new Element("abc"));
 *     vars.put("mynodes", new ParentNode[] {new Document(new Element("elem1")), new Element("elem2"))});
 *     xquery.execute(doc, null, vars);
 * </pre>
 * 
 * The <a target="_blank" href="http://www.saxonica.com/documentation/documentation.html">Standard XQuery functions</a> can be used directly.
 * Custom extension functions written in Java can be defined and used as explained in the
 * <a target="_blank" href="http://www.saxonica.com/documentation/extensibility/functions.html">Saxon Extensibility Functions</a> documentation.
 * <p>
 * Custom URI resolvers can be defined as explained in the
 * <a target="_blank" href=" http://www.saxonica.com/documentation/extensibility/uri-resolver.html">Saxon URI Resolver</a> documentation.
 * These are made available to the query by calling <code>setURIResolver()</code> 
 * on a {@link DynamicQueryContext} (per execution),
 * or on the configuration object of a {@link StaticQueryContext} (per query).
 * <p>
 * Performance note: For simple XPath expressions you can get a throughput of 
 * several thousand executions/sec for 10KB XHTML-like input documents, 
 * served from memory (commodity PC 2004, JDK 1.4.2).
 * Be aware that this is an example ballpark figure at best, because use cases, 
 * documents and the complexity of queries vary wildly in practise. 
 *
 * @author whoschek.AT.lbl.DOT.gov
 * @author $Author: hoschek3 $
 * @version $Revision: 1.29 $, $Date: 2004/10/27 18:50:30 $
 */
public class XQuery {
	
	private final XQueryExpression expression; // immutable hence implicitly thread-safe
	private final StaticQueryContext staticContext; // immutable hence implicitly thread-safe
	
	/*
	 * Native mode uses a DocumentWrapper ala net.sf.saxon.jdom, and is usually MANY
	 * times faster than usage of DOMSource. Native mode also requires
	 * dramatically less memory because it does not need to build a huge
	 * intermediary xerces DOM tree. This flag is an *undocumented* feature,
	 * mainly used for testing purposes, and for temporary work-arounds in case
	 * a bug in the wrappers should be reported.
	 */
	private static final boolean useNativeMode = 
		"true".equals(System.getProperty("nux.xom.xquery.XQuery.useNativeMode", "true"));
	
	private static final boolean DEBUG = false;

	/**
	 * Constructs a new compiled XQuery from the given query.
	 * 
	 * @param query
	 *            the query to compile
	 * @param baseURI
	 *            the base URI of the query. Used for resolving any relative
	 *            URI's found in the query, and used by the XQuery
	 *            <code>doc</code> function. (May be <code>null</code>
	 *            in which case it defaults to the current working directory).
	 * @throws XQueryException
	 *             if the syntax of the query is wrong, or if it references
	 *             namespaces, variables, or functions that have not been
	 *             declared, or contains other static errors.
	 */		
	public XQuery(String query, URI baseURI) throws XQueryException  {
		this(query, baseURI, null, null);
	}
	
	/**
	 * Constructs a new compiled XQuery from the given query, base URI, static
	 * context and resolver.
	 * 
	 * @param query
	 *            the query to compile
	 * @param baseURI
	 *            the base URI of the query. Used for resolving any relative
	 *            URI's found in the query, and used by the XQuery
	 *            <code>doc()</code> function, and hence the resolver. May be
	 *            <code>null</code> in which case it defaults to the current
	 *            working directory.
	 * @param staticContext
	 *            the context and configuration to use; per query (may be <code>null</code>).
	 * @param resolver
	 *            an object that is called by the XQuery processor to turn a URI
	 *            passed to the XQuery <code>doc()</code> function into a XOM
	 *            {@link Document}. May be <code>null</code> in which case
	 *            Saxon's default resolution is used.
	 * 
	 * @throws XQueryException
	 *             if the syntax of the query is wrong, or if it references
	 *             namespaces, variables, or functions that have not been
	 *             declared, or contains other static errors.
	 */
	public XQuery(String query, URI baseURI, StaticQueryContext staticContext, 
			final DocumentURIResolver resolver) throws XQueryException {
		
		if (query == null) throw new IllegalArgumentException("query must not be null");
		if (staticContext == null) staticContext = createStaticContext();
		this.staticContext = staticContext;
		if (baseURI != null) staticContext.setBaseURI(baseURI.toString());
		staticContext.getConfiguration().setHostLanguage(Configuration.XQUERY);
		
		if (resolver != null) {
			staticContext.getConfiguration().setURIResolver(
				new URIResolver() {
					public Source resolve(String href, String base) throws TransformerException {
						try {
							Document doc = resolver.resolve(href, base);
							if (doc != null) 
								return createNodeInfo(doc);
							else // let saxon try to resolve the URI itself
								return null;
						} catch (ValidityException e) {
							throw new TransformerException(e);
						} catch (ParsingException e) {
							throw new TransformerException(e);
						} catch (IOException e) {
							throw new TransformerException(e);
						}
					}
				}
			);
		}
		
		// generates saxon's compiled query representation
		try {
			this.expression = staticContext.compileQuery(query);
		} catch (TransformerException e) {
			throw new XQueryException(e);
		}
	}	
	
	/**
	 * Executes (evaluates) the query against the given node (subtree).
	 * 
	 * @param contextNode
	 *            the context node to execute the query against. The context
	 *            node is available to the query as the value of the query
	 *            expression ".". If this parameter is <code>null</code>, the
	 *            context node will be undefined.
	 * @return an iterator producing execution output results on demand
	 * @throws XQueryException
	 *             if an error occurs during execution
	 */
	public ResultSequence execute(ParentNode contextNode) throws XQueryException  {
		return execute(contextNode, null, null);
	}
	
	/**
	 * Executes (evaluates) the query against the given node (subtree), using the
	 * given dynamic context and external variables.
	 * <p>
	 * Argument <code>variables</code> specifies external global variables in the form of zero or more 
	 * <code>variableName --> variableValue</code> map associations.
	 * Each map entry's key and value are interpreted as follows:
	 * <ul>
	 *    <li>variableName (key): The name of the variable in "{uri}local-name" format.
     *     It is not an error to supply a value for a variable that has not been
     *     declared in the query, the variable will simply be ignored. If the variable has
     *     been declared in the query (as an external global variable) then it
     *     will be initialized with the value supplied.</li>
     * 
	 *    <li>variableValue (value): The value of the variable. This can be any valid Java
     *     object. In particular it can be a {@link ParentNode} object or a 
     *     {@link ParentNode}[] array.
     *     It follows the same conversion rules as a value returned
     *     from a Saxon extension function. An error will occur at query
     *     execution time if the supplied value cannot be converted to the required
     *     type as declared in the query. For precise control of the type of the
     *     value, instantiate one of the classes in the {@link net.sf.saxon.value} package,
     *     for example {@link net.sf.saxon.value.DateTimeValue}.</li>
	 * </ul>
	 * 
	 * @param contextNode
	 *            the context node to execute the query against. The context
	 *            node is available to the query as the value of the query
	 *            expression ".". If this parameter is <code>null</code>, the
	 *            context node will be undefined.
	 * @param dynamicContext
	 *            optional dynamic context of this execution (may be
	 *            <code>null</code>).
	 *            If not <code>null</code>, the Configuration object of the 
	 *            dynamic context must be be the same as the Configuration 
	 *            object that was used when creating the StaticQueryContext.
	 * @param variables
	 *            optional external global variables on the dynamic context; 
	 *            per execution (may be <code>null</code>).
	 * @return an iterator producing execution output results on demand
	 * @throws XQueryException
	 *             if an error occurs during execution
	 */
	public ResultSequence execute(ParentNode contextNode, 
			DynamicQueryContext dynamicContext, Map variables) throws XQueryException {
		
		if (dynamicContext == null)
			dynamicContext = new DynamicQueryContext(this.staticContext.getConfiguration());
		try {
			if (variables != null) 
				setVariables(dynamicContext, variables);
			if (contextNode != null)
				dynamicContext.setContextNode(createNodeInfo(contextNode));
		} catch (TransformerException e) {
			throw new XQueryException(e);
		}
		return newResultSequence(this.expression, dynamicContext);
	}
	
	/**
	 * Callback that returns a result sequence for the current query execution.
	 * <p>
	 * An XQuery result sequence may, apart from "normal" nodes, also contain
	 * <b>top-level</b> values of <a target="_blank"
	 * href="http://www.w3.org/TR/xpath-functions/#datatypes">atomic types </a>
	 * such as Long, Float, Boolean, etc, all of which are not XML nodes.
	 * <p>
	 * This method's result sequence implementation converts each <b>top-level</b>
	 * atomic value to an {@link Element} named "atomic" with a child
	 * {@link Text} node holding the value's string representation. 
	 * "Normal" nodes and anything not at top-level is returned "as is", without conversion.
	 * <p>
	 * Overrride this default implementation if you need custom conversions in
	 * result sequence implementations. Note however, that conversions of atomic
	 * values should rarely be used. It is often more desirable to avoid such
	 * atomic conversions altogether. This can almost always easily be achieved
	 * by formulating the <b>xquery string</b> such that it wraps atomic values via
	 * standard XQuery constructors (rather than via Java methods) into an element
	 * or attribute or document. That way, the query always produces a "normal"
	 * XML node sequence as output, and never produces a sequence of atomic
	 * values as output, and thus custom conversion by this class are never
	 * needed and invoked.
	 * <p>
	 * For example, by default the following query 
	 * <pre>
	 *     for $i in (1, 2, 3) return $i
	 * </pre>
	 * yields this (perhaps somewhat unexpected) output:
	 * <pre>
	 *     &lt;atomic type="java.lang.Long">1&lt;/atomic>
	 *     &lt;atomic type="java.lang.Long">2&lt;/atomic>
	 *     &lt;atomic type="java.lang.Long">3&lt;/atomic>
	 * </pre>
	 * <p>
	 * Hence, most likely you will want to rewrite the query along the following lines: 
	 * <pre>
	 *     for $i in (1, 2, 3) return &lt;item> {$i} &lt;/item>
	 * </pre>
	 * now yielding as output:
	 * <pre>
	 *     &lt;item>1&lt;/item> 
	 *     &lt;item>2&lt;/item>
	 *     &lt;item>3&lt;/item>
	 * </pre>
	 * Observe that the rewritten query converts top-level atomic values 
	 * precisely as desired by the user, without needing any Java-level conversion.
	 * 
	 * @param expression
	 *            the compiled query expression
	 * @param dynamicContext
	 *            the dynamic context of this execution
	 * @return a result sequence implementation
	 * @throws XQueryException
	 *             if an error occurs during execution
	 */
	protected ResultSequence newResultSequence(XQueryExpression expression, 
			DynamicQueryContext dynamicContext) throws XQueryException {
		
		try {
			return new DefaultResultSequence(expression.iterator(dynamicContext));
		} catch (TransformerException e) {
			throw new XQueryException(e);
		}
	}
	
	/** Returns a Node that saxon can read from. */
	private NodeInfo createNodeInfo(ParentNode node) throws TransformerException {
		if (node == null) throw new NullPointerException("node must not be null");
		if (!useNativeMode) 
			return this.staticContext.buildDocument(createSource(node));
		
		if (node instanceof Document) {
			return new DocumentWrapper((Document) node, node.getBaseURI(), staticContext.getConfiguration());
		}
		else if (node instanceof Element) {
			Element elem = (Element) node;
			Document doc = elem.getDocument();
			if (doc == null) doc = new Document(new Element("dummyRoot"));
			return new DocumentWrapper(doc, node.getBaseURI(), staticContext.getConfiguration()).wrap(elem);
		}
		else throw new IllegalArgumentException("unknown parent node type");
	}
	
	/** Returns a Source that saxon can read from. */
	private Source createSource(ParentNode node) {
		if (node == null) throw new NullPointerException("node must not be null");
		return createDOMSource(node);
		//return createXOMSource(node);
	}
	
	private Source createDOMSource(ParentNode node) {
		String systemID = node.getBaseURI();
		if (systemID != null && systemID.length() == 0) systemID = null; // probably safer that way !?
		
		org.w3c.dom.Node domNode;
		if (node instanceof Document) {
			domNode = DOMConverter.convert((Document) node, XOMUtil.getDOMImplementation());
		}
		else if (node instanceof Element) {
			// since node.copy() is expensive, it would be nice 
			// to have a method DOMConverter.convert(nu.xom.Element, DOMImplementation)
			Document doc = new Document((Element)node.copy()); // leave elem unmodified
			domNode = DOMConverter.convert(doc, XOMUtil.getDOMImplementation()).getDocumentElement();
		}
		else throw new IllegalArgumentException("unknown parent node type");
		
		return new DOMSource(domNode, systemID);
	}
		
	/** Setup variables of context, if any */
	private void setVariables(DynamicQueryContext dynamicContext, Map variables) 
			throws TransformerException {
		
		Iterator iter = variables.entrySet().iterator();
		while (iter.hasNext()) {
			Map.Entry entry = (Map.Entry) iter.next();
			if (entry.getKey() == null) throw new IllegalArgumentException("variable key must not be null");
			String name = entry.getKey().toString();
			Object value = entry.getValue();
			
			if (value instanceof ParentNode || value instanceof ParentNode[]) { 
				// pass xml document(s) such that saxon understands
				// available to query via
				// declare variable $foo as node() external;
				// or
				// declare variable $foo as node()* external;
				if (value instanceof ParentNode) 
					value = new ParentNode[] { (ParentNode) value };
				ParentNode[] nodes = (ParentNode[]) value;
				List sources = new ArrayList(nodes.length);
				for (int i = 0; i < nodes.length; i++) {
					if (useNativeMode) 
						sources.add(createNodeInfo(nodes[i]));
					else 
						sources.add(createSource(nodes[i]));
				}
				value = sources;
			}
			else if (value instanceof File || value instanceof File[]) {
				// undocumented feature (temporary?)
				// WARNING: Saxon's error messages do not well indicate what kind of 
				// well-formedness or validity problem a bad file has - if the file 
				// is bad you are left without much clue.
				//
				// pass xml document(s) such that saxon understands
				// available to query via
				// declare variable $foo as node() external;
				// or 
				// declare variable $foo as node()* external;
				if (value instanceof File) 
					value = new File[] { (File) value };
				File[] files = (File[]) value;
				List sources = new ArrayList(files.length);
				for (int i = 0; i < files.length; i++) {
					if (! files[i].exists()) 
						throw new TransformerException(
							new FileNotFoundException(files[i].toString()));
					if (files[i].isDirectory())
						throw new TransformerException(
							"File must not be a directory: " + files[i]);
					InputSource in = new InputSource(files[i].toURI().toString());
					sources.add(new SAXSource(
							staticContext.getConfiguration().getSourceParser(), in));
				}
				value = sources;
			}
			else if (value instanceof CharSequence) {
				value = new UntypedAtomicValue((CharSequence) value);
			}
			
			dynamicContext.setParameter(name, value);
		}
	}
	
	/** set defaults */
	private StaticQueryContext createStaticContext() {
		StaticQueryContext context = new StaticQueryContext(new Configuration());
		Configuration config = context.getConfiguration();
		config.setHostLanguage(Configuration.XQUERY);
		config.setErrorListener(new DefaultErrorListener());
		return context;
	}
	
	///////////////////////////////////////////////////////////////////////////////
	// Nested classes:
	///////////////////////////////////////////////////////////////////////////////
	
	private static class DefaultErrorListener implements ErrorListener {
		public void error(TransformerException e) throws TransformerException {
			throw e;
		}
		public void fatalError(TransformerException e) throws TransformerException {
			throw e;
		}
		public void warning(TransformerException e) throws TransformerException {
			System.err.println("Warning: " + e);
		}
	}
	
	/**
	 * An iterator representing an XQuery result sequence; allows to collect
	 * execution output in streaming or batched manner.
	 */
	private class DefaultResultSequence implements ResultSequence {
		
		// the underlying saxon iterator producing result items
		private final SequenceIterator results;
		
		private DefaultResultSequence(SequenceIterator results) {
			this.results = results;
		}
		
		public Node next() throws XQueryException {
			try {
				Node node;
				do {
					Item item = results.next();
					if (item == null) return null; // iterator exhausted
					node = convert(XPathEvaluator.convert(item));
				} while (node == null); // until convert() accepts an item
				return node;
			} catch (TransformerException e) {
				throw new XQueryException(e);
			}
		}
		
		 public Nodes toNodes() throws XQueryException {
			Nodes nodes = new Nodes();
			Node node;
			while ((node = next()) != null) {
				nodes.append(node);
			}
			return nodes;
		}
		
		/** Converts a saxon execution result node to XOM. */
		private Node convert(Object value) {
			if (DEBUG) System.err.println("value.getClass="+ value.getClass().getName());
			if (value instanceof Node) { 
				return (Node)value;
			}
			else if (value instanceof org.w3c.dom.Node) {
				if (value instanceof org.w3c.dom.Text) {						
					return DOMConverter.convert((org.w3c.dom.Text)value);
				}
				else if (value instanceof org.w3c.dom.Attr) {						
					return DOMConverter.convert((org.w3c.dom.Attr)value);
				}
				else if (value instanceof org.w3c.dom.Element) {
					return DOMConverter.convert((org.w3c.dom.Element)value);
				}
				else if (value instanceof org.w3c.dom.Document) {
					return DOMConverter.convert((org.w3c.dom.Document)value);
				}
				else if (value instanceof org.w3c.dom.Comment) {
					return DOMConverter.convert((org.w3c.dom.Comment)value);
				}
				else if (value instanceof org.w3c.dom.ProcessingInstruction) {
					return DOMConverter.convert((org.w3c.dom.ProcessingInstruction)value);
				}		
				else if (value instanceof org.w3c.dom.DocumentType) {
					return DOMConverter.convert((org.w3c.dom.DocumentType)value);
				}
				else throw new IllegalArgumentException(
					"unknown DOM node type - should never happen: " + value.getClass().getName());
			}
			else {
				return convertAtomicValue(value);
			}
		}

		/** Converts saxon's atomic value to XOM. */
		private Node convertAtomicValue(Object value) {
			if (DEBUG) System.err.println("atomicValue.getClass="+value.getClass().getName());
			Element elem = new Element("atomic");
			elem.addAttribute(new Attribute("type", value.getClass().getName()));
			elem.appendChild(value.toString());
			return elem;
		}

	}
	
// **********************************************************************
// Below is old garbage
// **********************************************************************
	
//	/**
//	 * Returns all remaining nodes from the result sequence, wrapped into a
//	 * single document and decorated Saxon-style.
//	 * 
//	 * @return a document
//	 * @throws TransformerException
//	 *             if an error occurs during execution
//	 */ 
//	public Document toDocument() throws TransformerException {
//		DocumentInfo output = QueryResult.wrap(results, staticContext.getConfiguration());
//		if (DEBUG) System.err.println(XMLUtil.toString((org.w3c.dom.Document) output));
//		return DOMConverter.convert((org.w3c.dom.Document) output);
//	}

}
